<?php
namespace Gath;

use Gath\Api\EndpointInterface;
use Gath\Api\EndpointPluginManager;
use Gath\Api\ResultSet;
use GuzzleHttp\Client as HttpClient;
use GuzzleHttp\ClientInterface as HttpClientInterface;
use GuzzleHttp\Exception\GuzzleException;
use GuzzleHttp\Psr7\Request;
use InvalidArgumentException;
use Psr\Http\Message\UriInterface;
use RuntimeException;
use Zend\Hydrator\DelegatingHydrator;
use Zend\Hydrator\HydratorInterface;
use Zend\Hydrator\HydratorPluginManager;
use Zend\Hydrator\Strategy\CollectionStrategy;
use Zend\Hydrator\Strategy\StrategyEnabledInterface;

/**
 * Gath klijent.
 *
 * @package Gath
 * @author Danijel Fabijan <hipnaba@gmail.com>
 * @link https://bitbucket.org/autokreso/gath
 */
final class Client implements ClientInterface
{
    /** @var UriInterface */
    private $uri;
    /** @var EndpointPluginManager */
    private $endpoints;
    /** @var HttpClientInterface */
    private $http;
    /** @var HydratorPluginManager */
    private $hydrators;

    /**
     * Konstruktor.
     *
     * @param UriInterface $uri
     * @param EndpointPluginManager $endpoints
     * @param HydratorPluginManager $hydrators
     */
    public function __construct(UriInterface $uri, EndpointPluginManager $endpoints, HydratorPluginManager $hydrators)
    {
        $this->uri = $uri;
        $this->endpoints = $endpoints;
        $this->hydrators = $hydrators;
    }

    /**
     * Vraća informacije o API-ju.
     *
     * @return Api\Info|Api\ResponseInterface
     */
    public function getApiInfo(): Api\Info
    {
        return $this->call('api-info');
    }

    /**
     * Poziva API endpoint koji vraća listu rezultata.
     *
     * @param $endpoint
     * @param array $params
     *
     * @return Api\ResultSetInterface
     */
    public function query($endpoint, array $params = []): Api\ResultSetInterface
    {
        if (!$endpoint instanceof EndpointInterface) {
            $endpoint = $this->getEndpoint($endpoint);
        }

        /** @var HydratorInterface $hydrator */
        $hydrator = $this->hydrators->get(ResultSet::class);

        if ($hydrator instanceof StrategyEnabledInterface) {
            $hydrator->addStrategy('results', new CollectionStrategy(
                $this->hydrators->get(DelegatingHydrator::class), $endpoint->getResultClass()
            ));
        }

        $data = $this->request($endpoint, $params);
        return $hydrator->hydrate($data, new ResultSet());
    }

    /**
     * @param EndpointInterface|string $endpoint
     * @param array $params
     * @return Api\ResponseInterface
     */
    public function call($endpoint, array $params = []): Api\ResponseInterface
    {
        if (!$endpoint instanceof EndpointInterface) {
            $endpoint = $this->getEndpoint($endpoint);
        }

        $data = $this->request($endpoint, $params);

        $responseClass = $endpoint->getResponseClass();
        /** @var Api\ResponseInterface $response */
        $response = new $responseClass();

        $hydrator = $this->hydrators->has($responseClass)
            ? $this->hydrators->get($responseClass)
            : $this->hydrators->get(DelegatingHydrator::class);

        return $hydrator->hydrate($data, $response);
    }

    /**
     * @param string|EndpointInterface $endpoint
     *
     * @param array $params
     * @param string $method
     * @return array
     */
    private function request($endpoint, array $params = [], string $method = 'GET'): array
    {
        if ($endpoint instanceof EndpointInterface) {
            $method = $endpoint->getMethod();
            $endpoint = $endpoint->getPath();
        }

        $uri = $this->uri->withPath($endpoint);

        if ('POST' === $method) {
            $params = ['json' => $params];
        } else if (!empty($params)) {
            $params = ['query' => $params];
        }

        try {
            $http = $this->getHttpClient();
            $response = $http->request($method, $uri, $params);
        } catch (GuzzleException $e) {
            throw new RuntimeException(
                sprintf(
                    "Došlo je do greške prilikom komunikacije s Gathom.\nUrl: %s\nUpit: %s",
                    (string) $uri,
                    json_encode($params)
                ),
                0,
                $e
            );
        }

        $contents = $response->getBody()->getContents();
        return json_decode($contents, true);
    }

    /**
     * @param string $name
     * @return EndpointInterface
     */
    private function getEndpoint(string $name): EndpointInterface
    {
        if (!$this->endpoints->has($name)) {
            throw new InvalidArgumentException(sprintf(
                "Nepoznati Gath endpoint '%s'.", $name
            ));
        }

        return $this->endpoints->get($name);
    }

    /**
     * @return HttpClientInterface
     */
    private function getHttpClient(): HttpClientInterface
    {
        if (null === $this->http) {
            $this->http = new HttpClient();
        }

        return $this->http;
    }
}
